from .connector import HassConnector, HassEvent, HassServiceCall
from .matcher import match_hass_state_changed, match_sunrise, match_sunset
from .skill import HassSkill
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import asyncio
import json
import logging
import urllib.parse

import aiohttp
from voluptuous import Required

from opsdroid.connector import Connector, register_event
from opsdroid.events import Event

_LOGGER = logging.getLogger(__name__)
CONFIG_SCHEMA = {
    Required("token"): str,
    Required("url"): str,
}


class HassEvent(Event):
    """Event class to represent a Home Assistant event."""


class HassServiceCall(Event):
    """Event class to represent making a service call in Home Assistant."""

    def __init__(self, domain, service, data, *args, **kwargs):
        self.domain = domain
        self.service = service
        self.data = data
        super().__init__(*args, **kwargs)


class HassConnector(Connector):
    """An opsdroid connector for syncing events with the Home Assistant event loop.

    """
    def __init__(self, config, opsdroid=None):
        super().__init__(config, opsdroid=opsdroid)
        self.name = "homeassistant"
        self.default_target = None
        self.connection = None
        self.listening = None
        self.discovery_info = None
        self.token = self.config.get("token")
        self.api_url = urllib.parse.urljoin(self.config.get("url"), "/api/")
        self.websocket_url = urllib.parse.urljoin(self.api_url, "websocket")
        self.id = 1

    def _get_next_id(self):
        self.id = self.id + 1
        return self.id

    async def connect(self):
        self.discovery_info = await self.query_api('discovery_info')
        self.listening = True

    async def listen(self):
        async with aiohttp.ClientSession() as session:
            while self.listening:
                try:
                    async with session.ws_connect(self.websocket_url) as ws:
                        self.connection = ws
                        async for msg in self.connection:
                            if msg.type == aiohttp.WSMsgType.TEXT:
                                await self._handle_message(json.loads(msg.data))
                            elif msg.type == aiohttp.WSMsgType.ERROR:
                                break
                    _LOGGER.info("Home Assistant closed the websocket, retrying...")
                except aiohttp.client_exceptions.ClientConnectorError:
                    _LOGGER.info("Unable to connect to Home Assistant, retrying...")
                await asyncio.sleep(1)

    async def query_api(self, endpoint, method='GET', decode_json=True, **params):
        """Query a Home Assistant API endpoint.

        The Home Assistant API can be queried at ``<hass url>/api/<endpoint>``. For a full reference
        of the available endpoints see https://developers.home-assistant.io/docs/en/external_api_rest.html.

        Args:
            endpoint: The endpoint that comes after /api/.
            method: HTTP method to use
            decode_json: Whether JSON should be decoded before returning
            **params: Parameters are specified as kwargs.
                      For GET requests these will be sent as url params.
                      For POST requests these will be dumped as a JSON dict and send at the post body.

        """
        url = urllib.parse.urljoin(self.api_url + '/', endpoint)
        headers = {
            "Authorization": "Bearer " + self.token,
            "Content-Type": "application/json"
        }
        _LOGGER.debug("Making a %s request to %s", method, url)
        async with aiohttp.ClientSession() as session:
            if method.upper() == 'GET':
                async with session.get(url, headers=headers, params=params) as resp:
                    if resp.status >= 400:
                        _LOGGER.error("Error %s - %s", resp.status, await resp.text())
                    else:
                        response = await resp.text()
            if method.upper() == 'POST':
                async with session.post(url, headers=headers, data=json.dumps(params)) as resp:
                    if resp.status >= 400:
                        _LOGGER.error("Error %s - %s", resp.status, await resp.text())
                    else:
                        response = await resp.text()
        if decode_json:
            response = json.loads(response)
        return response

    async def _handle_message(self, msg):
        msg_type = msg.get("type")

        if msg_type == "auth_required":
            await self.connection.send_json(
                {"type": "auth", "access_token": self.token}
            )

        if msg_type == "auth_invalid":
            _LOGGER.error("Invalid Home Assistant auth token.")
            await self.disconnect()

        if msg_type == "auth_ok":
            _LOGGER.info("Authenticated with Home Assistant.")
            await self.connection.send_json(
                {
                    "id": self._get_next_id(),
                    "type": "subscribe_events",
                    "event_type": "state_changed",
                }
            )

        if msg_type == "event":
            try:
                event = HassEvent(raw_event=msg)
                await event.update_entity("event_type", msg["event"]["event_type"])
                await event.update_entity("entity_id", msg["event"]["data"]["entity_id"])
                await event.update_entity(
                    "state", msg["event"]["data"]["new_state"]["state"]
                )
                changed = (
                    msg["event"]["data"]["old_state"] is None or
                    msg["event"]["data"]["new_state"]["state"]
                    != msg["event"]["data"]["old_state"]["state"]
                )
                await event.update_entity("changed", changed)
                await self.opsdroid.parse(event)
            except (TypeError, KeyError):
                _LOGGER.error("Home Assistant sent an event which didn't look like one we expected.")
                _LOGGER.error(msg)

        if msg_type == "result":
            if msg["success"]:
                pass
            else:
                _LOGGER.error("%s - %s", msg["error"]["code"], msg["error"]["message"])

    @register_event(HassServiceCall)
    async def send_service_call(self, event):
        await self.connection.send_json(
            {
                "id": self._get_next_id(),
                "type": "call_service",
                "domain": event.domain,
                "service": event.service,
                "service_data": event.data,
            }
        )

    async def disconnect(self):
        self.discovery_info = None
        self.listening = False
        await self.connection.close()



from typing import Callable

from opsdroid.matchers import match_event
from ..connector import HassEvent


def match_hass_state_changed(entity_id: str, **kwargs) -> Callable:
    """A matcher for state changes in Home Assistant.

    When an entity changes state in Home Assistant an event is triggered in Opsdroid.
    This matcher can be used to watch for a specific entity to change state::

        from opsdroid_homeassistant import HassSkill, match_hass_state_changed


        class SunriseSkill(HassSkill):

            @match_hass_state_changed("sun.sun", state="below_horizon")
            async def lights_on_at_sunset(self, event):
                await self.turn_on("light.outside")

    Alternatively you can use the :func:`opsdroid.matchers.match_event` matcher from the core Opsdroid set
    of matchers along with the :class:`opsdroid_homeassistant.HassEvent` class from opsdroid-homeassistant.
    With this method you much specify the entity with the kwarg ``entity_id`` and set ``changed=True`` if you
    only wish for the skill to trigger when the state changes. Sometimes entities send an event even if the
    state hasn't changed::

        from opsdroid.matchers import match_event
        from opsdroid_homeassistant import HassSkill, HassEvent


        class SunriseSkill(HassSkill):

            @match_event(HassEvent, entity_id="sun.sun", changed=True, state="below_horizon")
            async def lights_on_at_sunset(self, event):
                await self.turn_on("light.outside")

    Note:
        For sunrise and sunset triggers you can also use the :func:`match_sunrise` and
        :func:`match_sunset` helper matchers.

    Args:
        entity_id: The full domain and name of the entity you want to watch. e,g ``sun.sun``
        state (optional): The state you want to watch for. e.g ``on``

    """
    return match_event(HassEvent, entity_id=entity_id, changed=True, **kwargs)


match_sunrise = match_hass_state_changed("sun.sun", state="above_horizon")
match_sunrise.__doc__ = """A matcher to trigger skills on sunrise.

    This matcher can be used to run skills when the sun rises::

        from opsdroid_homeassistant import HassSkill, match_sunrise


        class SunriseSkill(HassSkill):

            @match_sunrise
            async def lights_off_at_sunrise(self, event):
                await self.turn_off("light.outside")

    """


match_sunset = match_hass_state_changed("sun.sun", state="below_horizon")
match_sunset.__doc__ = """A matcher to trigger skills on sunset.

    This matcher can be used to run skills when the sun sets::

        from opsdroid_homeassistant import HassSkill, match_sunset


        class SunsetSkill(HassSkill):

            @match_sunset
            async def lights_on_at_sunset(self, event):
                await self.turn_on("light.outside")

    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import os

import versioneer

HERE = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(HERE, "README.md"), "r", encoding="utf8") as fh:
    README = fh.read()
with open(os.path.join(HERE, "requirements.txt"), "r") as fh:
    REQUIRES = [line.strip() for line in fh]

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Jacob Tomlinson",
    author_email="jacob@tomlinson.email",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Home Assistant support for opsdroid",
    install_requires=REQUIRES,
    license="Apache Software License 2.0",
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="Opsdroid Home Assistant",
    name="opsdroid-homeassistant",
    packages=["opsdroid_homeassistant"],
    entry_points={
        "opsdroid_connectors": ["homeassistant = opsdroid_homeassistant.connector"]
    },
    url="https://github.com/opsdroid/opsdroid-homeassistant",
    zip_safe=False,
)



import pytest

from opsdroid_homeassistant import HassConnector


@pytest.fixture
def dummy_config():
    return {"token": "abc123", "url": "https://example.com:8123"}


def test_attributes(dummy_config):
    connector = HassConnector(dummy_config, opsdroid=None)

    assert connector.name == "homeassistant"
    assert connector.default_target is None



import arrow
from datetime import datetime
import logging

from opsdroid.skill import Skill

from ..connector import HassServiceCall

_LOGGER = logging.getLogger(__name__)


class HassSkill(Skill):
    """An Opsdroid skill base class with Home Assistant helper methods.

    To create a Home Assistant skill for your Opsdroid bot you need to subclass
    this class and define methods which are decorated with matchers. For triggering skills on
    Home Assistant events you will likely want to use the
    :func:`opsdroid_homeassistant.match_hass_state_changed` matcher.

    This base class also has some helper functions such as :meth:`turn_on` and :meth:`turn_off`
    to make it quick and convenient to control your Home Assistant entities. See the methods
    below for more information.

    Examples:

       Turning a light on at sunset::

            from opsdroid_homeassistant import HassSkill, match_hass_state_changed


            class SunriseSkill(HassSkill):

                @match_hass_state_changed("sun.sun", state="below_horizon")
                async def lights_on_at_sunset(self, event):
                    await self.turn_on("light.outside")

    For detailed information on writing skills check out the Opsdroid skill docs
    https://docs.opsdroid.dev/en/stable/skills/index.html.

    """

    def __init__(self, opsdroid, config, *args, **kwargs):
        super().__init__(opsdroid, config, *args, **kwargs)
        [self.hass] = [
            connector
            for connector in self.opsdroid.connectors
            if connector.name == "homeassistant"
        ]

    async def call_service(self, domain: str, service: str, *args, **kwargs):
        """Send a service call to Home Assistant.

        Build your own service call to any domain and service.

        Note:
            For common operations such and turning off and on entities see
            the :meth:`turn_on` and :meth:`turn_off` helper functions.
        """
        await self.hass.send(HassServiceCall(domain, service, kwargs))

    async def get_state(self, entity: str):
        """Get the state of an entity.

        Args:
            entity: The ID of the entity to get the state for.

        Returns:
            The state of the entity.

        Examples:
            Get the state of the sun sensor::

                >>> await self.get_state("sun.sun")
                "above_horizon"
        """
        state = await self.hass.query_api("states/" + entity)
        _LOGGER.debug(state)
        return state.get("state", None)

    async def turn_on(self, entity_id: str, **kwargs):
        """Turn on an entity in Home Assistant.

        Sends a ``homeassistant.turn_on`` service call with the specified entity.
        """
        await self.call_service(
            "homeassistant", "turn_on", entity_id=entity_id, **kwargs
        )

    async def turn_off(self, entity_id: str, **kwargs):
        """Turn off an entity in Home Assistant.

        Sends a ``homeassistant.turn_off`` service call with the specified entity.
        """
        await self.call_service(
            "homeassistant", "turn_off", entity_id=entity_id, **kwargs
        )

    async def toggle(self, entity_id: str, **kwargs):
        """Toggle an entity in Home Assistant.

        Sends a ``homeassistant.toggle`` service call with the specified entity.
        """
        await self.call_service(
            "homeassistant", "toggle", entity_id=entity_id, **kwargs
        )

    async def update_entity(self, entity_id: str, **kwargs):
        """Request an entity update in Home Assistant.

        Sends a ``homeassistant.update_entity`` service call with the specified entity.
        """
        await self.call_service(
            "homeassistant", "update_entity", entity_id=entity_id, **kwargs
        )

    async def set_value(self, entity_id: str, value, **kwargs):
        """Sets an entity to the specified value in Home Assistant.

        Depending on the entity type provided one of the following services will be called:

        * ``input_number.set_value``
        * ``input_text.set_value``
        * ``input_select.select_option``

        """
        if "input_number" in entity_id:
            await self.call_service(
                "input_number", "set_value", entity_id=entity_id, value=value, **kwargs
            )
        elif "input_text" in entity_id:
            await self.call_service(
                "input_text", "set_value", entity_id=entity_id, value=value, **kwargs
            )
        elif "input_select" in entity_id:
            await self.call_service(
                "input_select",
                "select_option",
                entity_id=entity_id,
                option=value,
                **kwargs
            )
        else:
            _LOGGER.error("Unable to set value for unsupported entity %s", entity_id)

    async def notify(
        self, message: str, title="Home Assistant", target="notify", **kwargs
    ):
        """Send a notification to Home Assistant.

        Sends a ``notify.notify`` service call with the specified title and message.

        Args:
            message: A message to notify with.
            title (optional): A title to set in the notification.
            target (optional): The notification target. Defaults to ``notify`` which will notify all.
        """
        await self.call_service(
            "notify", target, message=message, title=title, **kwargs
        )

    async def sun_up(self):
        """Check whether the sun is up.

        Returns:
            True if sun is up, else False.

        """
        sun_state = await self.get_state("sun.sun")
        return sun_state == "above_horizon"

    async def sun_down(self):
        """Check whether the sun is down.

        Returns:
            True if sun is down, else False.

        """
        sun_state = await self.get_state("sun.sun")
        return sun_state == "below_horizon"

    async def sunrise(self):
        """Get the timestamp for the next sunrise.

        Returns:
            A Datetime object of next sunrise.

        """
        sun_state = await self.hass.query_api("states/sun.sun")
        sunrise = arrow.get(sun_state["attributes"]["next_rising"])
        return sunrise.datetime

    async def sunset(self):
        """Get the timestamp for the next sunset.

        Returns:
            A Datetime object of next sunset.

        """
        sun_state = await self.hass.query_api("states/sun.sun")
        sunset = arrow.get(sun_state["attributes"]["next_setting"])
        return sunset.datetime

    async def get_trackers(self):
        """Get a list of tracker entities from Home Assistant.

        Returns:
            List of tracker dictionary objects.

        Examples:

            >>> await self.get_trackers()
            [{
                "attributes": {
                    "entity_picture": "https://www.gravatar.com/avatar/00000000000000000000000000000000?s=500&d=mm",
                    "friendly_name": "jacob",
                    "ip": "192.168.0.2",
                    "scanner": "NmapDeviceScanner",
                    "source_type": "router"
                },
                "context": {
                    "id": "abc123",
                    "parent_id": None,
                    "user_id": None
                },
                "entity_id": "device_tracker.jacob",
                "last_changed": "2020-01-03T20:27:55.001812+00:00",
                "last_updated": "2020-01-03T20:27:55.001812+00:00",
                "state": "home"
            }]

        """
        states = await self.hass.query_api("states")
        device_trackers = [
            entity
            for entity in states
            if entity["entity_id"].startswith("device_tracker.")
        ]
        return device_trackers

    async def anyone_home(self):
        """Check if anyone is home.

        Returns:
            True if any tracker is set to ``home``, else False.

        """
        trackers = await self.get_trackers()
        home = [tracker["state"] == "home" for tracker in trackers]
        return any(home)

    async def everyone_home(self):
        """Check if everyone is home.

        Returns:
            True if all trackers are set to ``home``, else False.

        """
        trackers = await self.get_trackers()
        home = [tracker["state"] == "home" for tracker in trackers]
        return all(home)

    async def nobody_home(self):
        """Check if nobody is home.

        Returns:
            True if all trackers are set to ``not_home``, else False.

        """
        trackers = await self.get_trackers()
        not_home = [tracker["state"] == "not_home" for tracker in trackers]
        return all(not_home)

    async def render_template(self, template: str) -> str:
        """Ask Home Assistant to render a template.

        Home Assistant has a built in templating engine powered by Jinja2.

        https://www.home-assistant.io/docs/configuration/templating/

        This method allows you to pass a template string to Home Assistant and
        it will return the formatted response.

        Args:
            template: The template string to be rendered by Home Assistant.

        Returns:
            A formatted string of the template.

        Examples:

            >>> await self.render_template("Jacob is at {{ states('device_tracker.jacob') }}!")
            Jacob is at home!

        """
        return await self.hass.query_api(
            "template", method="POST", decode_json=False, template=template
        )
